/* 
 *  SSE3DNow Copyright (c) Roy Longbottom 2002 - 2004
 *
 *  Can be compiled with Microsoft Visual C++ 6.0 Pro with Service 
 *  Pack 5 and Processor Pack, required to use the new instructions.
 *  These are used via assembly code or intrinsic functions. The
 *  Processor Pack includes a help file, procpack.chm, that describes
 *  all the intrinsics. For full details of Intel assembly instructions
 *  see intel.com:
 *
 *   IA-32 Intel Architecture Software Developer's Manual Volume 2: 
 *   Instruction Set Reference.
 *
 *  This program only uses the simpler floating point instructions. There are 
 *  others for SSE, SSE2 and 3DNow and non-floating point instructions. 
 *
 *  SSE and SSE2 use 128 bit registers in Single Instruction Multiple Data SIMD 
 *  mode. SSE is via 4 single precision 32 bit numbers and SSE2 via 2 double 
 *  precision 64 bit numbers. 3DNow uses two 32 bit numbers in MMX registers. 
 *  SSE instructions also need Operating System support to control their use, 
 *  so OS support has also to be determined.
 *
 *
 *  What the benchmark does
 *
 *  Calculates s=s+x[m]*y[m] and x[m]=x[m]+y[m] using SSE, SSE2 and
 *  3DNow instructions, as available, plus normal Single Precision
 *  FP instructions. Tests are carried out with different memory use
 *  from 4 Kbytes upwards.
 *
 *  Data is loaded in unrolled loops 64 bytes at a time to suit 64 byte bus 
 *  burst reads, leading to the number of instructions in a loop varying 
 *  between SSE, SSE2, 3DNow and SPFP. Obtaining high performance using 
 *  normal floating point instructions is rather difficult with multiples 
 *  of eight loads, calculations and stores. So this has been left for the 
 *  compiler to do. 
 *
 *  Modules
 *
 *  Config.cpp - via WhatConfig() identifies CPU type/speed, supported
 *             instruction extensions, memory size and OS version.
 *
 *  SSE3DNow.rc - menus and dialog boxes, AboutDlgProc(), TestPanel(),
 *        RunPanel().
 *
 *        TestPanel shows configuration details on starting and results
 *        at the end of testing. It has Run Tests and Exit buttons.
 *        RunPanel provides a speed summary as tests are running. 
 *
 *  SSE3DNow.h and SSE3DNow.ico
 *
 *  SSE3DNow.cpp
 *
 *  Particularly with Windows XP, on running a benchmark which displays
 *  on-going results, the window for the latter can be covered by another
 *  window. To overcome this, all tests are run via the WinMain message
 *  processing loop using phase and jval variables, with jval controlling
 *  the size of memory to be tested.
 *
 *  Phase 1 via RUN_ALL button and MENU_RUNALL executes setUp() to 
 *  allocate maximum memory and open the results log file SSE3DNow.txt.
 *
 *  Phase 2 to 9 executes runTests() jval times, each for a particular
 *  runTime. The tests are only run if the appropriate instructions
 *  are available.
 *
 *  Phase 2 - SSE2One() assembly code
 *  Phase 3 - SSEOne() assembly code
 *  Phase 4 - One3DNow() assembly code
 *  Phase 5 - SPOne C code loop
 *  Phase 6 - SSE2Two() assembly code
 *  Phase 7 - SSETwo() assembly code
 *  Phase 8 - Two3DNow() assembly code
 *  Phase 9 - SPTwo C code loop
 *
 *  Current results are displayed at the start of runTests() and the log
 *  file is updated after phase 9.
 *
 *  Each do while (secs < runTime) loop executes checkTime()to increase
 *  the number of times each test is executed if needed (more testing
 *  less overheads).
 *
 *  start_time() and end_time() call high resolution timer to obtain
 *  running time secs.
 *  local_time() obtains date and clock time
 *  finish()completes the log file and releases allocated memory.
 *  displayWhatever() either closes TestPanel or closes RunPanel and
 *  opens TestPanel.
 *
 *  InitApp() determines maximum memory size to be used (25% of RAM ) but 
 *  this can be increased by commandLine parameters KBytes where test 
 *  runTime can also be changed from default of 0.1 seconds and autoRun
 *  set to run, log results and exit automatically.
 */
 
#define NAME "Test"
#define TITLE " Test Memory"
#define INITGUID

#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <malloc.h>
#include <xmmintrin.h>
#include "SSE3DNow.h"

double * xd;
double * yd;
float  * xs;
float  * ys;
double * dsum;
float  * ssum;

FILE    *outfile;

inline void SSE2One(int kd)
{
    __asm
    {
       mov     edx, kd
       mov     eax, xd
       mov     ecx, yd
       mov     ebx, dsum
       mov     edi,0                  
       movapd  xmm4, XMMWORD PTR [ebx]
   lp: movapd  xmm0, XMMWORD PTR [eax+edi]
       movapd  xmm1, XMMWORD PTR [eax+edi+16]
       movapd  xmm2, XMMWORD PTR [eax+edi+32]
       movapd  xmm3, XMMWORD PTR [eax+edi+48]
       mulpd   xmm0, XMMWORD PTR [ecx+edi]
       mulpd   xmm1, XMMWORD PTR [ecx+edi+16]
       mulpd   xmm2, XMMWORD PTR [ecx+edi+32]
       mulpd   xmm3, XMMWORD PTR [ecx+edi+48]
       add     edi,64               
       addpd   xmm4, xmm0
       addpd   xmm4, xmm1
       addpd   xmm4, xmm2
       addpd   xmm4, xmm3
       sub     edx,8                                    
       jg      lp                   
       movapd  XMMWORD PTR [ebx], xmm4
    }
}                         

inline void SSEOne(int ks)
{
    __asm
    {
       mov     edx, ks
       mov     eax, xd
       mov     ecx, yd
       mov     ebx, dsum
       mov     edi,0                  
       movaps  xmm7, XMMWORD PTR [ebx]
   lp: movaps  xmm0, XMMWORD PTR [eax+edi]
       movaps  xmm1, XMMWORD PTR [eax+edi+16]
       movaps  xmm2, XMMWORD PTR [eax+edi+32]
       movaps  xmm3, XMMWORD PTR [eax+edi+48]
       mulps   xmm0, XMMWORD PTR [ecx+edi]
       mulps   xmm1, XMMWORD PTR [ecx+edi+16]
       mulps   xmm2, XMMWORD PTR [ecx+edi+32]
       mulps   xmm3, XMMWORD PTR [ecx+edi+48]
       add     edi,64               
       addps   xmm7, xmm0
       addps   xmm7, xmm1
       addps   xmm7, xmm2
       addps   xmm7, xmm3
       sub     edx,16                
       jg      lp                   
       movaps  XMMWORD PTR [ebx], xmm7
    }
}



inline void SSE2Two(int kd)
{
    __asm
    {
       mov     edx, kd
       mov     eax, xd
       mov     ecx, yd
       mov     edi, 0
   lp: movapd  xmm0, XMMWORD PTR [eax+edi]
       movapd  xmm1, XMMWORD PTR [eax+edi+16]
       movapd  xmm2, XMMWORD PTR [eax+edi+32]
       movapd  xmm3, XMMWORD PTR [eax+edi+48]
       addpd   xmm0, XMMWORD PTR [ecx+edi]
       addpd   xmm1, XMMWORD PTR [ecx+edi+16]
       addpd   xmm2, XMMWORD PTR [ecx+edi+32]
       addpd   xmm3, XMMWORD PTR [ecx+edi+48]
       movapd  XMMWORD PTR [eax+edi], xmm0
       movapd  XMMWORD PTR [eax+edi+16], xmm1
       movapd  XMMWORD PTR [eax+edi+32], xmm2
       movapd  XMMWORD PTR [eax+edi+48], xmm3
       add     edi,64       
       sub     edx,8        
       jg      lp                        
    }    
}

inline void SSETwo(int ks)
{
    __asm
    {
       mov     edx, ks
       mov     eax, xd
       mov     ecx, yd
       mov     edi, 0
   lp: movaps  xmm0, XMMWORD PTR [eax+edi]
       movaps  xmm1, XMMWORD PTR [eax+edi+16]
       movaps  xmm2, XMMWORD PTR [eax+edi+32]
       movaps  xmm3, XMMWORD PTR [eax+edi+48]
       addps   xmm0, XMMWORD PTR [ecx+edi]
       addps   xmm1, XMMWORD PTR [ecx+edi+16]
       addps   xmm2, XMMWORD PTR [ecx+edi+32]
       addps   xmm3, XMMWORD PTR [ecx+edi+48]
       movaps  XMMWORD PTR [eax+edi], xmm0
       movaps  XMMWORD PTR [eax+edi+16], xmm1
       movaps  XMMWORD PTR [eax+edi+32], xmm2
       movaps  XMMWORD PTR [eax+edi+48], xmm3
       add     edi,64       
       sub     edx,16        
       jg      lp
    }
}




inline void One3DNow(int ks)
{
    __asm
    {
       femms
       mov     edx, ks
       mov     eax, xd
       mov     ecx, yd
       mov     ebx, dsum
       mov     edi,0
       align   16
   lp: movq    mm0, [eax+edi]
       movq    mm1, [eax+edi+8]
       movq    mm2, [eax+edi+16]
       movq    mm3, [eax+edi+24]
       movq    mm4, [eax+edi+32]
       movq    mm5, [eax+edi+40]
       movq    mm6, [eax+edi+48]
       movq    mm7, [eax+edi+56]
       pfmul   mm0, [ecx+edi]
       pfmul   mm1, [ecx+edi+8]
       pfmul   mm2, [ecx+edi+16]
       pfmul   mm3, [ecx+edi+24]
       pfmul   mm4, [ecx+edi+32]
       pfmul   mm5, [ecx+edi+40]
       pfmul   mm6, [ecx+edi+48]
       pfmul   mm7, [ecx+edi+56]
       add     edi,64               
       pfadd   mm0, mm1
       pfadd   mm0, mm2
       pfadd   mm0, mm3
       pfadd   mm0, mm4
       pfadd   mm0, mm5
       pfadd   mm0, mm6
       pfadd   mm0, mm7
       movq    mm7, [ebx]
       pfadd   mm7, mm0
       movq    [ebx], mm7
       sub     edx, 16                
       jg      lp                   
       femms
    }
}

inline void Two3DNow(int ks)
{
    __asm
    {
       femms
       mov     edx, ks
       mov     eax, xd
       mov     ecx, yd
       mov     edi, 0
       align   16
   lp: movq    mm0, [eax+edi]
       movq    mm1, [eax+edi+8]
       movq    mm2, [eax+edi+16]
       movq    mm3, [eax+edi+24]
       movq    mm4, [eax+edi+32]
       movq    mm5, [eax+edi+40]
       movq    mm6, [eax+edi+48]
       movq    mm7, [eax+edi+56]
       pfadd   mm0, [ecx+edi]
       movq    [eax+edi],    mm0
       pfadd   mm1, [ecx+edi+8]
       movq    [eax+edi+8],  mm1
       pfadd   mm2, [ecx+edi+16]
       movq    [eax+edi+16], mm2
       pfadd   mm3, [ecx+edi+24]
       movq    [eax+edi+24], mm3
       pfadd   mm4, [ecx+edi+32]
       movq    [eax+edi+32], mm4
       pfadd   mm5, [ecx+edi+40]
       movq    [eax+edi+40], mm5
       pfadd   mm6, [ecx+edi+48]
       movq    [eax+edi+48], mm6
       pfadd   mm7, [ecx+edi+56]
       movq    [eax+edi+56], mm7
       add     edi, 64       
       sub     edx, 16        
       jg      lp
       femms
    }
}


//  ************** PROTOPTYPES *****************

BOOL FAR PASCAL AboutDlgProc( HWND, UINT, WPARAM, LPARAM);
static BOOL InitApp(HINSTANCE, LPSTR, int);
long FAR PASCAL WindowProc(HWND, UINT, WPARAM, LPARAM);                  
static void CleanUpAndPostQuit(void);
void displayWhatever(HWND);

BOOL FAR PASCAL TestPanel( HWND, UINT, WPARAM, LPARAM);
BOOL FAR PASCAL RunPanel( HWND, UINT, WPARAM, LPARAM);

void local_time();
void end_time();
void start_time();
void checkTime();
BOOL runTests();
BOOL setUp();
BOOL finish();

// ************* GLOBAL VARIABLES **********
 
char   version[30] = "Version 1.0";

char    timeday[30];
LARGE_INTEGER starttime;
double  secs;
int     n1;
double  runTime = 0.1;

BOOL bQuit = FALSE;

HWND   mainHWND;
HWND   testHwnd = NULL;
HWND   runHwnd = NULL;
HINSTANCE MyInstance;  // HANDLE 
char    windowText[256];
char    commandLine[1024] = "";

short     keyPress;
BOOL  autoRun = FALSE;
BOOL  doTesting = FALSE;
char   writeMsg[256];
char   writeMsg1[256];
char   writeMsg2[256];
char   writeMsg3[256];
char   writeMsg4[256];
double results[30][9];
double maxmbps[8];

int    memUsed[100];
int  passesDone;
BOOL  runDone = FALSE;
int  ramKBused;
int  ramKB;
int runs = 10; // 2 MB;
int vector[20];
int allocMem[20];
RECT     winRect;
int     screenWide;
int     screenHigh;

int    maxrows = 13;
int    userows;
double hmfp, hmint;
double hmfp2, hmint2;
double sdfp, sdint;
double sdfp2, sdint2;
char  startRam[200];
int useMem;
int phase = 0;
int jval = 0;   


void displayWhatever(HWND hwnd)
{
    sprintf(windowText, " SSE, SSE2 and 3DNow Memory Test");
    SetWindowText(mainHWND, windowText);        
        
    if (doTesting)
    {
        if (testHwnd)   DestroyWindow(testHwnd);
    }
    else
    {    
        if (runHwnd)    DestroyWindow(runHwnd);
        testHwnd = CreateDialog( MyInstance,"DIALOG_8", hwnd, (DLGPROC)TestPanel);
    }  
} // displayWhatever


// WinMain - initialization, message loop

int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR cmdline, int cmdshow)
{
    MSG         msg;
    
    hPrevInstance = hPrevInstance;

    if (!InitApp(hInstance, cmdline, cmdshow))
         return FALSE;

    if (autoRun)
    {
         phase = 1;
    }        
    while(!bQuit)
    {
        if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) )
        {
            
           if (msg.message == WM_QUIT)
           {
                CleanUpAndPostQuit();
                break;
           }
   
           if( !GetMessage( &msg, NULL, 0, 0 ) ) return msg.wParam;
           TranslateMessage(&msg);
           DispatchMessage(&msg);
        }
        else if (phase == 1)
        {
            if (!setUp())
            {
                CleanUpAndPostQuit();
                phase = 0;
            }
            phase = 2;
            jval = 0;
        }
        else if (phase > 1)
        {           
            if (!runTests())
            {
                CleanUpAndPostQuit();
                phase = -1;
            }
            phase = phase + 1;

            if (phase == 10)
            {
                phase = 2;
                jval = jval + 1;
                if (jval == runs)
                {
                    phase = 0;
                    finish();
                    if (autoRun)
                    {
                        CleanUpAndPostQuit();
                    }
                }
            }        
        }
        else
        {
           // make sure we go to sleep if we have nothing else to do
           WaitMessage();
        }
    }
    return msg.wParam;
    
} // WinMain

// ***************************************************************************
//                    Initialization and object creation                     *
// ***************************************************************************

// InitApp
// Creates window

static BOOL
InitApp(HINSTANCE hInstance, LPSTR cmdline, int cmdshow)
{
    HWND hwnd;
    WNDCLASS wc;
    int   mult = 20;
    int   i, j;
    char *delims = {" ,"};
    char commands[20][256];
    char *p;
    int  KBytes;
    char buf[256] = "";
    HDC  hdc;
    int msecs = 0;
  
    
    // set up and registers the window class
    
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = sizeof(DWORD);
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(hInstance, "GenericIcon");
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName = "GenericMenu";
    wc.lpszClassName = NAME;
    if (!RegisterClass(&wc))
        return FALSE;

    // Create the window

    hwnd =
        CreateWindow(
            NAME,
            TITLE,
            WS_OVERLAPPEDWINDOW | WS_HSCROLL | WS_VSCROLL,
            0,
            0,
            640,
            480,
            NULL, 
            NULL, 
            hInstance, 
            NULL 
        );

    if (!hwnd) return FALSE;

    mainHWND = hwnd;
    strcpy(commandLine, strupr(cmdline));

    hdc = GetDC(mainHWND);
    screenWide = GetDeviceCaps(hdc, HORZRES);
    screenHigh = GetDeviceCaps(hdc, VERTRES);   
    ReleaseDC(mainHWND, hdc);

    winRect.left = 0;
    winRect.top = 0;
    winRect.right = screenWide;
    winRect.bottom = screenHigh;

    WhatConfig();
    strcpy(startRam, configData3);
    ramKB = ramsize / 1024;
    strcpy(writeMsg1, configData1);
    strcpy(writeMsg2, configData5);
    strcpy(writeMsg3, configData2);
    strcpy(writeMsg4, configData4);

    allocMem[0] = 2048;  // two arrays
    for (i=1; i<19; i++)
    {
        allocMem[i] = allocMem[i-1] * 2;
    }
    vector[0] = 240;       //      4 KB
    vector[1] = 496;       //      8 KB
    vector[2] = 1008;      //     16 KB 
    vector[3] = 2000;      //     32 KB
    vector[4] = 4000;      //     64 KB
    vector[5] = 8000;      //    128 KB
    vector[6] = 16000;     //    256 KB
    vector[7] = 32000;     //    512 KB
    vector[8] = 64000;     //    1 MB
    vector[9] = 128000;    //    2 MB
    vector[10] = 256000;   //    4 MB
    vector[11] = 512000;   //    8 MB
    vector[12] = 1024000;  //   16 MB
    vector[13] = 2048000;  //   32 MB
    vector[14] = 4096000;  //   64 MB
    vector[15] = 8192000;  //  128 MB
    vector[16] = 16384000; //  256 MB
    vector[17] = 32768000; //  512 MB
    vector[18] = 65536000; // 1024 MB

    if (ramKB >   14000) runs = 11;
    if (ramKB >   30000) runs = 12;
    if (ramKB >   60000) runs = 13;
    if (ramKB >  120000) runs = 14;
    if (ramKB >  250000) runs = 15;
    if (ramKB >  500000) runs = 16;
    if (ramKB > 1000000) runs = 17;
    if (ramKB > 2000000) runs = 18;
    if (ramKB > 4000000) runs = 19;


    i = 0;
    p = strtok(commandLine, delims);
    while (p != NULL)
    {
        strcpy(commands[i], p);
        strcat(buf, commands[i]);
        strcat(buf, " ");
        i = i + 1;
        if (i > 4) break;
        p = strtok(NULL, delims);
    }
    if (i > 0)
    {
        strcpy(commandLine, buf);
    }
    else
    {
        strcpy(commandLine, "None");
    }
    autoRun = FALSE;
    KBytes = 0;
       
    for (j=0; j<i; j++)
    {
        switch(commands[j][0])
        {            
            case 'A': // Auto Run
             autoRun = TRUE;
             break;            

            case 'K': // KBytes
             if (j < i -1)
             {
                 sscanf(commands[j+1], "%d", &KBytes);
             }
             break;            
       
            case 'M': // millisecs
             if (j < i -1)
             {
                 sscanf(commands[j+1], "%d", &msecs);
             }
             break;
        }            
    }
    if (KBytes > 0)
    {
        runs = 1;
        j = 8;
        for (i=0; i<18; i++)
        {
            if (KBytes > j - 1) runs = runs + 1;
            j = j * 2; 
        }
    }
    if (msecs > 0) runTime = ((double) msecs) / 1000;

    ramKBused = vector[runs-1] * 16 / 1000;

    keyPress = GetAsyncKeyState(VK_ESCAPE);

    // Display the window
    
    ShowWindow(hwnd, cmdshow);    
    UpdateWindow(hwnd);

    if (!autoRun) displayWhatever(mainHWND);
    
    return TRUE;
    
}  // InitApp



// WindowProc - handle messages for the main application window


long FAR PASCAL WindowProc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam)                  
{
           
    switch( msg )
    {
    case WM_ACTIVATEAPP:
        break;
        
    case WM_INITMENUPOPUP:
        break;
        
    case WM_CREATE:
        break;

    case WM_DESTROY:
        CleanUpAndPostQuit();
        break;

    case WM_COMMAND:
        switch( LOWORD( wparam ) )
        {
        case MENU_ABOUT:
            if (testHwnd)   DestroyWindow(testHwnd);
            DialogBox( MyInstance,"AboutBox", hwnd, (DLGPROC)AboutDlgProc );
            displayWhatever(mainHWND);                                      
            break;

       case MENU_RUNALL:
            doTesting = TRUE;
            displayWhatever(mainHWND);                                      
            phase = 1;
            break;


        case MENU_DISPLAY:
            displayWhatever(mainHWND);                                      
            break;
        
        case MENU_EXIT:
            CleanUpAndPostQuit();
            break;
                
        }
        break;
   
        case WM_PAINT:
             break;
           
        case WM_KEYDOWN:
            break;

       case WM_ACTIVATE:
            break;
    } 
    return DefWindowProc( hwnd, msg, wparam, lparam );

} // WindowProc




// AboutDlgProc - processes messages for the about dialog.

BOOL FAR PASCAL AboutDlgProc( HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam )
{
    lparam = lparam;

    switch( msg ) {
    case WM_INITDIALOG:
        return( TRUE );

    case WM_COMMAND:
        if( LOWORD( wparam ) == IDOK ) {
            EndDialog( hwnd, TRUE );
            return( TRUE );
        }
        break;
    }
    return( FALSE );

} // AboutDlgProc



  
void local_time()
{
    time_t time_of_day;
    time_of_day = time( NULL ); 
    strcpy (timeday, ctime(&time_of_day)); // timeday = date and time
    return;
}




// CleanUpAndPostQuit
// Release all D3DRM objects, post a quit message and set the bQuit flag

void
CleanUpAndPostQuit(void)
{
    bQuit = TRUE;
    return;
}


BOOL FAR PASCAL RunPanel( HWND hdlg, UINT msg,
                                WPARAM wparam, LPARAM lparam )
{
    wparam = wparam;
    lparam = lparam;
    hdlg = hdlg;
    int  k;
    int FAR  tabs[10] = {1, 30, 60, 90, 120, 150, 180, 210, 240, 270};

     switch( msg )
     {
       case WM_INITDIALOG:
       
        SendDlgItemMessage(hdlg, RESULTS1, LB_SETTABSTOPS, (WPARAM) 10, (LPARAM) (int FAR*) tabs); 
        sprintf(writeMsg, "Maximum RAM size to be tested %d KB", ramKBused);
        SendDlgItemMessage(hdlg, NOTES1, LB_ADDSTRING, 0, (LPARAM) writeMsg);
        sprintf(writeMsg, "Flashing disk light suggest insufficient RAM and paging activity "
                      "to/from RAM and disk.");
        SendDlgItemMessage(hdlg, NOTES1, LB_ADDSTRING, 0, (LPARAM) writeMsg);
        sprintf(writeMsg, "Press ESC to cancel and exit program. It might take some time.");
        SendDlgItemMessage(hdlg, NOTES1, LB_ADDSTRING, 0, (LPARAM) writeMsg);
       
        if (runDone)
        {
            SendDlgItemMessage(hdlg, RESULTS1, LB_SETHORIZONTALEXTENT, 700, 0L);
            sprintf(writeMsg, " SSE & 3DNow Test %s by Roy Longbottom"
                              "%10.3f seconds minimum per test", version, runTime);
            SendDlgItemMessage(hdlg, RESULTS1, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, " ");
            SendDlgItemMessage(hdlg, RESULTS1, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, "Memory%c  s=s+x[m]*y[m]  %c%c%c x[m]=x[m]+y[m]", 9, 9, 9, 9); 

            SendDlgItemMessage(hdlg, RESULTS1, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, "KBytes%c SSE2%c  SSE%c  3DN%c Sngl%c SSE2%c  SSE%c  3DN%c Sngl",
                                     9, 9, 9, 9, 9, 9, 9, 9);   
            SendDlgItemMessage(hdlg, RESULTS1, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, "  Used%c MB/S%c MB/S%c MB/S%c MB/S%c MB/S%c MB/S%c MB/S%c MB/S%",
                                     9, 9, 9, 9, 9, 9, 9, 9);           
            SendDlgItemMessage(hdlg, RESULTS1, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, " ");
            SendDlgItemMessage(hdlg, RESULTS1, LB_ADDSTRING, 0, (LPARAM) writeMsg);

            for (k=0; k<passesDone; k++)
            {
               sprintf(writeMsg, "%6d %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f",  
                                memUsed[k]/1000, 9, results[k][0], 9, results[k][1], 9, results[k][2],
                                                 9, results[k][3], 9, results[k][4], 9, results[k][5],
                                                 9, results[k][6], 9, results[k][7]);
               SendDlgItemMessage(hdlg, RESULTS1, LB_ADDSTRING, 0, (LPARAM) writeMsg);
               SendDlgItemMessage(hdlg, RESULTS1, LB_SETCURSEL, k+6, 0L);
            }
        }
        
        return( TRUE );              
     }
     return( FALSE );
    
} // runPanel

BOOL FAR PASCAL TestPanel( HWND hdlg, UINT msg,
                                WPARAM wparam, LPARAM lparam )
{
    wparam = wparam;
    lparam = lparam;
    hdlg = hdlg;
    int  k;
    int FAR  tabs[10] = {1, 30, 60, 90, 120, 150, 180, 210, 240, 270};

     switch( msg )
     {
       case WM_INITDIALOG:
        SendDlgItemMessage(hdlg, RESULTS, LB_SETTABSTOPS, (WPARAM) 10, (LPARAM) (int FAR*) tabs); 
        sprintf(writeMsg, "Maximum RAM size to be tested %d KB at %10.3f seconds per test.",
                                    ramKBused, runTime);
        SendDlgItemMessage(hdlg, NOTES, LB_ADDSTRING, 0, (LPARAM) writeMsg);
        sprintf(writeMsg, "RAM size from GlobalMemoryStatus %d KB", ramKB);
        SendDlgItemMessage(hdlg, NOTES, LB_ADDSTRING, 0, (LPARAM) writeMsg);
        sprintf(writeMsg, "After running, results should be in file SSE3DNow.txt and above");
        SendDlgItemMessage(hdlg, NOTES, LB_ADDSTRING, 0, (LPARAM) writeMsg);
       
        if (runDone)
        {
            SendDlgItemMessage(hdlg, RESULTS, LB_SETHORIZONTALEXTENT, 700, 0L);
            sprintf(writeMsg, " SSE & 3DNow Test %s by Roy Longbottom"
                              "%10.3f seconds minimum per test", version, runTime);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, " ");
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);

            sprintf(writeMsg, "Memory%c  s=s+x[m]*y[m]  %c%c%c x[m]=x[m]+y[m]%c%c", 9, 9, 9, 9); 
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, "KBytes%c SSE2%c  SSE%c  3DN%c Sngl%c SSE2%c  SSE%c  3DN%c Sngl",
                                     9, 9, 9, 9, 9, 9, 9, 9);   
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, "  Used%c MB/S%c MB/S%c MB/S%c MB/S%c MB/S%c MB/S%c MB/S%c MB/S",
                                     9, 9, 9, 9, 9, 9, 9, 9);           
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, " ");
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);

            for (k=0; k<passesDone; k++)
            {
               sprintf(writeMsg, "%6d %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f",  
                                memUsed[k]/1000, 9, results[k][0], 9, results[k][1], 9, results[k][2],
                                                 9, results[k][3], 9, results[k][4], 9, results[k][5],
                                                 9, results[k][6], 9, results[k][7]);
               SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            }
            sprintf(writeMsg, " ");
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, " Maximum");
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            sprintf(writeMsg, " Mflops%c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f %c%7.0f",
                    9, maxmbps[0], 9, maxmbps[1], 9, maxmbps[2], 9, maxmbps[3],
                    9, maxmbps[4], 9, maxmbps[5], 9, maxmbps[6], 9, maxmbps[7]);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);       
            SendDlgItemMessage(hdlg, RESULTS, LB_SETCURSEL, passesDone+11, 0L);

        }
        else
        {
            strcpy(writeMsg, "Batch Command ");
            strncat(writeMsg, commandLine, 200);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg1);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg2);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg3);
            SendDlgItemMessage(hdlg, RESULTS, LB_ADDSTRING, 0, (LPARAM) writeMsg4);
                }
        return( TRUE );
                    
       case WM_COMMAND:

        if( HIWORD( wparam ) == BN_CLICKED)
        {
            switch (LOWORD (wparam))
            {
                case RUN_ALL:
                  PostMessage(mainHWND, WM_COMMAND, MENU_RUNALL, 0);
                 break;

                case EXIT_BUTTON:
                 PostMessage(mainHWND, WM_COMMAND, MENU_EXIT, 0);
                 break;
            }
        }
        return( TRUE );         
     }
     return( FALSE );
    
} // TestPanel


void start_time()
{
    QueryPerformanceCounter(&starttime);
}
 
void end_time()
{
    LARGE_INTEGER liDiff;
    LARGE_INTEGER liFreq;

    QueryPerformanceCounter(&liDiff);

    liDiff.QuadPart -= starttime.QuadPart;

    (void)QueryPerformanceFrequency(&liFreq);
    
    secs = (double)liDiff.QuadPart / (double) liFreq.QuadPart;
}

BOOL setUp()
{
    int i, j;
    HDC  hdc;
    
    for (i=0; i<30; i++)
    {
        for (j=0; j<9; j++)
        {
            results[i][j] = 0;
        }
    }
    useMem = allocMem[runs - 1]; // vector[runs - 1] * 8 + 8;    

    hdc = GetDC(mainHWND);
    sprintf(writeMsg, "                          Allocating Memory  ");
    TextOut(hdc, 100, 100, writeMsg, lstrlen(writeMsg));    
    ReleaseDC(mainHWND, hdc);

    outfile = fopen("SSE3DNow.txt","a+");
    if (outfile == NULL)
    {
        sprintf(writeMsg, "Cannot open results file");
        if (!autoRun) MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
        CleanUpAndPostQuit();
        return FALSE;
    }

    local_time();
    fprintf(outfile, "\n ***************************************************************");
    fprintf(outfile, "\n   SSE, SSE2 and 3DNow Memory Reading Speed Test %s\n", version);      
    fprintf(outfile, "\n             Copyright (C) Roy Longbottom 2002\n");    
    fprintf(outfile, "\n %10.3f seconds per test, Start %s\n", runTime, timeday);
    fprintf(outfile, " MB/S is read speed. Multiply second set by 1.5 for read/write speed\n\n");
    fprintf(outfile, "  Memory          s=s+x[m]*y[m]               x[m]=x[m]+y[m]\n"); 
    fflush(outfile);

/*
  For the fastest SSE instructions used, all data loaded or stored MUST be 
  aligned on 16 byte address boundaries, otherwise the program hangs.

  useMem+65536 avoids severe performance problems with P4 CPU when 2 arrays
  are aligned exactly on MB boundaries
*/

    xd = (double *)_aligned_malloc(useMem+65536, 16);
    if (xd == NULL)
    {
        sprintf(writeMsg, "Cannot allocate memory");
        if (!autoRun) MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
        fprintf(outfile,"Cannot allocate memory\n\n");
        fflush(outfile);
        fclose(outfile);
        CleanUpAndPostQuit();
        return FALSE;
    }        
    yd = (double *)_aligned_malloc(useMem, 16);
    if (yd == NULL)
    {
        sprintf(writeMsg, "Cannot allocate memory");
        if (!autoRun) MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
        fprintf(outfile,"Cannot allocate memory\n\n");
        fflush(outfile);
        fclose(outfile);
        _aligned_free(xd);
        CleanUpAndPostQuit();
        return FALSE;
    }        
    dsum = (double *)_aligned_malloc(64, 16);
    if (dsum == NULL)
    {
        sprintf(writeMsg, "Cannot allocate memory");
        if (!autoRun) MessageBox(NULL, writeMsg, "ERROR WILL EXIT", MB_ICONWARNING | MB_OK);
        fprintf(outfile,"Cannot allocate memory\n\n");
        fflush(outfile);
        fclose(outfile);
        _aligned_free(xd);
        _aligned_free(yd);
        CleanUpAndPostQuit();
        return FALSE;
    }        

    xs   = (float  * )xd;
    ys   = (float  * )yd;
    ssum = (float  * )dsum;

    fprintf(outfile, "  KBytes    SSE2    SSE  3DNow   Sngl   SSE2    SSE  3DNow   Sngl\n");   
    fflush(outfile);

    runDone = TRUE;    
    passesDone = 0;

    sprintf(writeMsg, "                           Running Tests                    ");
    hdc = GetDC(mainHWND);
    TextOut(hdc, 100, 100, writeMsg, lstrlen(writeMsg));    
    ReleaseDC(mainHWND, hdc);
    GdiFlush();

    fprintf(outfile, "   Used     MB/S   MB/S   MB/S   MB/S   MB/S   MB/S   MB/S   MB/S\n\n");
    fflush(outfile);

    for (i=0; i<8; i++)
    {
        maxmbps[i] = 0;
    }
    return TRUE;

} // setUp


BOOL runTests()
{

    float  sums;
    float  ones = 1.0;
    float  zeros = 0;
    int i, m, j, nn, mem;
    int kd, ks;
    int inc;
    double memBytes;
    double BytesPerFLOP;
    double sumd;
    
    double oned = 1.0;
    double zerod = 0;
        
    if (runHwnd)    DestroyWindow(runHwnd);
    runHwnd = CreateDialog( MyInstance,"DIALOG_9", mainHWND, (DLGPROC)RunPanel);

    j = jval;
    {
        kd = vector[j];
        nn = 6400000 / kd;  // min 240
        if (nn < 1) nn = 1;
        
        ks = kd * 2;
        
        mem = kd * 16;
        memUsed[j] = mem + 200;
        
        inc = 4;

        // SSE2One *******************************  

        if (hasSSE2 && hasSSE2OS && phase == 2)
        {
            n1 = nn;
      
            do
            {
                sumd = 0;
                dsum[0] = 0;
                dsum[1] = 0;
                for (m=0; m<kd+6; m++)
                {
                      xd[m] = oned;
                      yd[m] = oned;
                }          
        
                start_time();
    
                for (i=0; i<n1; i++)
                {
                     SSE2One(kd);
                }
                end_time();
                checkTime();
            }
            while (secs < runTime);
    
            sumd = dsum[0] + dsum[1];
            memBytes = 16;
            results[j][0] = (double)kd * (double)n1 * memBytes / 1000000/ secs;
            BytesPerFLOP = 8;            
            if (results[j][0] / BytesPerFLOP > maxmbps[0]) maxmbps[0] = results[j][0] / BytesPerFLOP;
        }
    
        //  SSEOne ******************************* 

        if (hasSSE && hasSSEOS && phase == 3)
        {
            n1 = nn;
            do
              {
                ssum[0] = 0;
                ssum[1] = 0;
                ssum[2] = 0;
                ssum[3] = 0;
                for (m=0; m<ks+12; m++)
                {
                      xs[m] = ones;
                      ys[m] = ones;
                }          
                start_time();
                
                for (i=0; i<n1; i++)
                  {
                     SSEOne(ks);
                  }
                end_time();
                checkTime();            
              }
            while (secs < runTime);
    
            sums = ssum[0] + ssum[1] + ssum[2] + ssum[3];        
            memBytes = 8;
            results[j][1] = (double)ks * (double)n1 * memBytes / 1000000 / secs;
            BytesPerFLOP = 4;            
            if (results[j][1] / BytesPerFLOP > maxmbps[1]) maxmbps[1] = results[j][1] / BytesPerFLOP;
        }

        //  3DNowOne ******************************* 

        if (has3DNow && phase == 4)
        {
            n1 = nn;
            do
            {
                ssum[0] = 0;
                ssum[1] = 0;
                for (m=0; m<ks+12; m++)
                {
                      xs[m] = ones;
                      ys[m] = ones;
                }          
                start_time();
                
                for (i=0; i<n1; i++)
                {
                     One3DNow(ks);
                }
                end_time();
                checkTime();            
            }
            while (secs < runTime);
    
            sums = ssum[0] + ssum[1];        
            memBytes = 8;
            results[j][2] = (double)ks * (double)n1 * memBytes / 1000000 / secs;
            BytesPerFLOP = 4;            
            if (results[j][2] / BytesPerFLOP > maxmbps[2]) maxmbps[2] = results[j][2] / BytesPerFLOP;
        }

        //  SPOne ******************************* 

        if (phase == 5)
        {
            n1 = nn;
            inc = 16;
            do
              {
                ssum[0] = 0;
                for (m=0; m<ks; m++)
                  {
                      xs[m] = ones;
                      ys[m] = ones;
                  }          
                start_time();
                
                for (i=0; i<n1; i++)
                  {
                    for (m=0; m<ks-inc; m=m+inc)
                     {
                       sums = sums + xs[m] * ys[m]
                                   + xs[m+1]  * ys[m+1]
                                   + xs[m+2]  * ys[m+2]
                                   + xs[m+3]  * ys[m+3]                   
                                   + xs[m+4]  * ys[m+4]
                                   + xs[m+5]  * ys[m+5]
                                   + xs[m+6]  * ys[m+6]                   
                                   + xs[m+7]  * ys[m+7]
                                   + xs[m+8]  * ys[m+8]
                                   + xs[m+9]  * ys[m+9]                   
                                   + xs[m+10] * ys[m+10]
                                   + xs[m+11] * ys[m+11]
                                   + xs[m+12] * ys[m+12]                   
                                   + xs[m+13] * ys[m+13]
                                   + xs[m+14] * ys[m+14]
                                   + xs[m+15] * ys[m+15];                   
                     }
                  }
                end_time();
                checkTime();            
              }
            while (secs < runTime);
    
            if (sums == 0.0) fprintf(outfile, " SPOne Wrong Result\n");         
            memBytes = 8;
            results[j][3] = (double)ks * (double)n1 * memBytes / 1000000 / secs;
            BytesPerFLOP = 4;            
            if (results[j][3] / BytesPerFLOP > maxmbps[3]) maxmbps[3] = results[j][3] / BytesPerFLOP;
        }
    
        //  SSE2Two ******************************* 
        
        if (hasSSE2 && hasSSE2OS && phase == 6)
        {
            n1 = nn;
            do
              {
                for (m=0; m<kd; m++)
                  {
                      xd[m] = zerod;
                      yd[m] = oned;
                  }          
                start_time();
                for (i=0; i<n1; i++)
                  {
                     SSE2Two(kd);
                  }
                end_time();
                checkTime();
              }
            while (secs < runTime);
        
            sumd = xd[1];
            memBytes = 16;
            results[j][4] = (double)kd * (double)n1 * memBytes / 1000000 / secs;
            BytesPerFLOP = 16;            
            if (results[j][4] / BytesPerFLOP > maxmbps[4]) maxmbps[4] = results[j][4] / BytesPerFLOP;
        }
    
        //  SSETwo ******************************* 

        if (hasSSE && hasSSEOS && phase == 7)
        {
            n1 = nn;
            do
              {
                for (m=0; m<ks; m++)
                  {
                      xs[m] = zeros;
                      ys[m] = ones;
                  }          
                start_time();
                for (i=0; i<n1; i++)
                  {
                     SSETwo(ks);
                  }
                end_time();
                checkTime();
              }
            while (secs < runTime);
                
            sums = xs[1];
            memBytes = 8;
            results[j][5] = (double)ks * (double)n1 * memBytes / 1000000 / secs;
            BytesPerFLOP = 8;            
            if (results[j][5] / BytesPerFLOP > maxmbps[5]) maxmbps[5] = results[j][5] / BytesPerFLOP;
        }

        //  3DNowTwo ******************************* 

        if (has3DNow && phase == 8)
        {
            n1 = nn;
            do
              {
                for (m=0; m<ks; m++)
                  {
                      xs[m] = zeros;
                      ys[m] = ones;
                  }          
                start_time();
                for (i=0; i<n1; i++)
                  {
                     Two3DNow(ks);
                  }
                end_time();
                checkTime();
              }
            while (secs < runTime);
                
            sums = xs[1];
            memBytes = 8;
            results[j][6] = (double)ks * (double)n1 * memBytes / 1000000 / secs;
            BytesPerFLOP = 8;            
            if (results[j][6] / BytesPerFLOP > maxmbps[6]) maxmbps[6] = results[j][6] / BytesPerFLOP;
        }

        //  SPTwo ******************************* 

        if (phase == 9)
        {
            n1 = nn;
            inc = 16;
            do
              {
                for (m=0; m<ks; m++)
                  {
                      xs[m] = zeros;
                      ys[m] = ones;
                  }          
                start_time();
                for (i=0; i<n1; i++)
                  {
                    for (m=0; m<ks-inc; m=m+inc)
                     {
                          xs[m]    = xs[m]    + ys[m];
                          xs[m+1]  = xs[m+1]  + ys[m+1];
                          xs[m+2]  = xs[m+2]  + ys[m+2];
                          xs[m+3]  = xs[m+3]  + ys[m+3];
                          xs[m+4]  = xs[m+4]  + ys[m+4];
                          xs[m+5]  = xs[m+5]  + ys[m+5];
                          xs[m+6]  = xs[m+6]  + ys[m+6];
                          xs[m+7]  = xs[m+7]  + ys[m+7];
                          xs[m+8]  = xs[m+8]  + ys[m+8];
                          xs[m+9]  = xs[m+9]  + ys[m+9];
                          xs[m+10] = xs[m+10] + ys[m+10];
                          xs[m+11] = xs[m+11] + ys[m+11];
                          xs[m+12] = xs[m+12] + ys[m+12];
                          xs[m+13] = xs[m+13] + ys[m+13];
                          xs[m+14] = xs[m+14] + ys[m+14];
                          xs[m+15] = xs[m+15] + ys[m+15];
                     }
                  }
                end_time();
                checkTime();
              }
            while (secs < runTime);
                
            sums = xs[1];
            if (sums == 0.0) fprintf(outfile, " SPTwo Wrong Result\n");         
            memBytes = 8;
            results[j][7] = (double)ks * (double)n1 * memBytes / 1000000 / secs;
            BytesPerFLOP = 8;            
            if (results[j][7] / BytesPerFLOP > maxmbps[7]) maxmbps[7] = results[j][7] / BytesPerFLOP;

            fprintf(outfile, "%7d  %7.0f%7.0f%7.0f", memUsed[j] /1000,
                                                  results[j][0], results[j][1], results[j][2]);
            fprintf(outfile, "%7.0f%7.0f%7.0f",   results[j][3], results[j][4], results[j][5]);
            fprintf(outfile, "%7.0f%7.0f\n",      results[j][6], results[j][7]);
    
            fflush(outfile);
        }
        passesDone = j + 1;
        if (runHwnd)    DestroyWindow(runHwnd);
    
        if (GetAsyncKeyState(VK_ESCAPE) != keyPress)
        {
            fprintf(outfile, "\n\n\n");
            _aligned_free(xd);
            _aligned_free(yd);
            _aligned_free(dsum);
            fflush(outfile);
            fclose(outfile);
            CleanUpAndPostQuit();
            return FALSE;
        }
    }
    return TRUE;
    
} // runTests

BOOL finish()
{
    int i;
    HDC  hdc;
    
    fprintf(outfile, "\n");
    fprintf(outfile, "            SSE2    SSE  3DNow   Norm   SSE2    SSE  3DNow   Norm\n");   
    fprintf(outfile, " Maximum      DP     SP     SP     SP     DP     SP     SP     SP\n");
    fprintf(outfile, "  MFLOPS ");
    for (i=0; i<8; i++)
    {
        fprintf(outfile, "%7.0f", maxmbps[i]);
    }
    fprintf(outfile, "\n");
    

    local_time();
    if (autoRun) DestroyWindow(runHwnd);

    fprintf(outfile, "\n                End of test %s", timeday);        
    fprintf(outfile,"\n");

    WhatConfig();
    fprintf(outfile,"%s\n", configData1);
    if (hasSSEOS || hasSSE2OS) fprintf(outfile,"%s\n", configData5);
    fprintf(outfile,"%s\n", configData2);
    if (hasMMX || hasSSE || hasSSE2) fprintf(outfile,"%s\n", configData4);
    fprintf(outfile,"%s, at end %d KB\n", startRam, freesize / 1024);        
    fprintf(outfile,"\n\n");

    fflush(outfile);
    fclose(outfile);
    _aligned_free(xd);
    _aligned_free(yd);
    _aligned_free(dsum);
    doTesting = FALSE;
    if (!autoRun) displayWhatever(mainHWND);                                      

    hdc = GetDC(mainHWND);
    Rectangle(hdc, winRect.left, winRect.top,
                   winRect.right, winRect.bottom);
    ReleaseDC(mainHWND, hdc);

    return TRUE;
    
} // finish


void checkTime()
{
    if (secs < runTime)
    {
          if (secs < runTime / 8.0)
          {
                n1 = n1 * 10;
          }
          else
          {
                n1 = (int)(runTime * 1.25 / secs * (double)n1+1);
           }
      }
}


         
